#[derive(Debug)]
pub enum Instruction {
  IncrementPointer,
  DecrementPointer,
  IncrementValue,
  DecrementValue,
  OutputValue,
  ReadValue,
  Branch(Vec<Instruction>),
}

struct Input<'a> {
  source: &'a str,
  position: usize,
}

impl Input<'_> {
  fn read_all(&mut self) -> Vec<Instruction> {
    let mut result: Vec<Instruction> = vec!();
    while !self.done_reading() {
      if let Some(next) = self.read_next() {
        result.push(next);
      } else {
        break;
      }
    }
    result
  }

  fn read_next(&mut self) -> Option<Instruction> {
    while !self.done_reading() && self.next_is_comment() {
      self.skip_next_char();
    }
    if self.peek_char() == Some('[') {
      return self.read_branch();
    } else {
      return self.read_instruction_char()
    }
  }

  fn read_branch(&mut self) -> Option<Instruction> {
    let mut branch: Vec<Instruction> = vec!();
    self.skip_next_char(); // [
    while self.peek_char() != Some(']') {
      if self.next_is_comment() {
        self.skip_next_char()
      } else if let Some(next) = self.read_next() {
        branch.push(next);
      } else {
        return None;
      }
    }
    self.skip_next_char(); // ]
    Some(Instruction::Branch(branch))
  }

  fn peek_char(&self) -> Option<char> {
    self.source.chars().collect::<Vec<_>>().get(self.position).copied()
  }

  fn next_is_comment(&self) -> bool {
    if let Some(next) = self.peek_char() {
      !vec!['>', '<', '+', '-', '.', ',', '['].contains(&next)
    } else {
      false
    }
  }

  fn read_instruction_char(&mut self) -> Option<Instruction> {
    let c = self.peek_char();
    self.skip_next_char();
    match c {
      Some('>') => Some(Instruction::IncrementPointer),
      Some('<') => Some(Instruction::DecrementPointer),
      Some('+') => Some(Instruction::IncrementValue),
      Some('-') => Some(Instruction::DecrementValue),
      Some('.') => Some(Instruction::OutputValue),
      Some(',') => Some(Instruction::ReadValue),
      _ => None,
    }
  }

  fn done_reading(&self) -> bool {
    self.position >= self.source.chars().count()
  }

  fn skip_next_char(&mut self) {
    if !self.done_reading() {
      self.position += 1;
    }
  }
}

pub fn parse(source: &str) -> Vec<Instruction> {
  let mut input = Input { source, position: 0 };
  input.read_all()
}
